//@ts-check
import * as Y from 'yjs'
import { redo, undo, ySyncPlugin, yUndoPlugin } from 'y-prosemirror'
import { EditorState } from 'prosemirror-state'
import { EditorView } from 'prosemirror-view'
import { keymap } from 'prosemirror-keymap'
import { schema } from './schema'
import getProsemirrorSetup from './prosemirror-setup'
import './styles/main.css'
import './styles/style.css'
import { serializeUpdate, deserializeUpdate } from 'webxdc-yjs-provider';
import { WebxdcSyncProvider } from 'webxdc-yjs-provider/initWebxdcSyncProviderGeneric';
import throttle from 'just-throttle'
/** @typedef {import('../webxdc.d.ts')} */

// Currently there is no UI to toggle this.
const manualSend = localStorage.getItem('manualSend') !== 'false';
const alwaysSendOnClose = localStorage.getItem('alwaysSendOnClose') !== 'false';

const ydoc = new Y.Doc()
/** @type {ReturnType<typeof createEditor>} */
let prosemirror
const type = ydoc.getXmlFragment('prosemirror')

let sentEditsButNotNotifiedPeersAboutItYet = false;
/** @type {undefined | string} */
let lastSetWebxdcDocumentName = undefined;
/** @type {undefined | string} */
let lastSetWebxdcSummary = undefined;

const provider = new WebxdcSyncProvider(
  ydoc,
  serializeUpdate,
  deserializeUpdate,
  outgoingSerializedYjsUpdate => {
    /** @type {Parameters<typeof webxdc.sendUpdate>[0]} */
    const update = {
      payload: {
        serializedYjsUpdate: outgoingSerializedYjsUpdate,
      },
    };

    const preferredDocumentName =
      prosemirror?.state.doc.content.firstChild?.textContent
      .trimStart()
      .slice(0, 100)
      .trimEnd();
    if (
      preferredDocumentName
      && preferredDocumentName !== lastSetWebxdcDocumentName
    ) {
      update.document = preferredDocumentName;
      lastSetWebxdcDocumentName = preferredDocumentName;
    }

    // TODO i18n
    const preferredSummary = `Last edit: ${webxdc.selfName}`;
    if (lastSetWebxdcSummary !== preferredSummary) {
      update.summary = preferredSummary;
    }

    webxdc.sendUpdate(update, 'Document edited');
    updateSyncButtonState();
    sentEditsButNotNotifiedPeersAboutItYet = true;
  }
);
const storedUpdatesAppliedPromise = webxdc.setUpdateListener(update => {
  if (update.payload?.serializedYjsUpdate) {
    provider.onIncomingYjsUpdate(update.payload.serializedYjsUpdate);
  }
  if (update.document) {
    lastSetWebxdcDocumentName = update.document;
  }
  if (update.summary) {
    lastSetWebxdcSummary = update.summary;
  }
});
{
let onNeedToSendLocalUpdates;
if (manualSend) {
  // Updates are to be sent with the "sync" button.
  onNeedToSendLocalUpdates = () => {
    updateSyncButtonState();
  };
} else {
  /**
   * Delta Chat Core throttles updates pretty heavily:
   * https://github.com/deltachat/deltachat-core-rust/blob/b96028cd87f02a83f8f0a5282da4b4bb88cdc05c/src/context.rs#L379
   * https://codeberg.org/webxdc/editor/pulls/23#issuecomment-996164
   * So we better spend the quota wisely, i.e. not spend all of it immediately
   * and then get throttled by the messenger implementation.
   */
  const throttlePeriodMs = 10 * 1000;
  const throttledSendLocalUpdates = throttle(
    () => provider.sendUnsentLocalUpdates(),
    throttlePeriodMs,
    // TODO maybe use both leading and trailing.
    // But it doesn't work properly in `just-throttle` IMO:
    // https://github.com/angus-c/just/issues/207#issuecomment-1621879811
    { leading: false, trailing: true }
  );
  onNeedToSendLocalUpdates = () => {
    throttledSendLocalUpdates();
    updateSyncButtonState();
  }
}
provider.onNeedToSendLocalUpdates = onNeedToSendLocalUpdates;
}

addEventListener("DOMContentLoaded", () => {
  /** @type {HTMLDivElement} */
  const editor = document.querySelector('#editor')
  prosemirror = createEditor(editor);

  if (alwaysSendOnClose || !manualSend) {
    // TODO improvement: detach the listener when there are no pending changes?
    // Or is that unnecessary for this event? I heard stuff about
    // back/forward cache.
    window.addEventListener("visibilitychange", () => {
      if (document.visibilityState === "hidden") {
        sendUnsentUpdatesAndNotifyPeersAboutLastEdit();
      }
    });

    // TODO only add this listener when there are unsaved changes (see
    // the comment right above).
    window.addEventListener('beforeunload', sendUpdatesAndDelayClose);
  }
  // TODO maybe add `else` with `beforeunload` listener that prevents the close
  // if there are unsaved changes.

  initExtraButtons();
})

// TODO fix: even if the user undid all of their changes (Ctrl+Z),
// we still send a message that they "made an edit".
//
// Keep in mind:
// - If we didn't make edits to the document, there is no need to send
//   a notification.
// - If the updates were sent already (say, with a save button), but we haven't
//   sent a notification about them yet, we need to do so eventually.
// - If we sent a notification about our edits, and then made new edits,
//   we need to send a new notification.
function sendUnsentUpdatesAndNotifyPeersAboutLastEdit() {
  provider.sendUnsentLocalUpdates();

  if (sentEditsButNotNotifiedPeersAboutItYet) {
    // Could be an empty string, or `undefined`
    const docText = lastSetWebxdcDocumentName != undefined
      ? `'${lastSetWebxdcDocumentName}'`
      : 'the document';
    // TODO perf: batch the Y.Doc updates and this in a single `sendUpdate`
    // call. Do we use `provider.takeOutUnsentLocalUpdates` + a wrapper
    // `sendUpdate` function to achieve this?
    webxdc.sendUpdate({
      payload: null,
      // TODO i18n.
      info: `${webxdc.selfName} edited ${docText}`,
    }, 'Notify chat members about edits');
    sentEditsButNotNotifiedPeersAboutItYet = false;
  }
}

// A workaround for
// https://github.com/deltachat/deltachat-desktop/issues/3321#issue-1814659377
// https://codeberg.org/webxdc/editor/issues/12#issuecomment-989270
let pendingClose = false;
/**
 * @param {BeforeUnloadEvent} e
 */
function sendUpdatesAndDelayClose(e) {
  sendUnsentUpdatesAndNotifyPeersAboutLastEdit();
  if (!pendingClose) {
    pendingClose = true;
    setTimeout(() => {
      window.removeEventListener('beforeunload', sendUpdatesAndDelayClose);
      // Yes, we close the parent. This relies on the fact that we run inside an
      // `<iframe>`, which is the case for Delta Chat Desktop.
      try {
        window.parent.close();
      } catch (e) {
        window.close();
      }
    }, 100);
  }

  e.preventDefault();
  // For compatibility with older browsers
  e.returnValue = '';
  return '';
}

/**
 * @param {HTMLElement} element
 */
function createEditor(element) {
  return new EditorView(element, {
    state: EditorState.create({
      schema,
      // TODO refactor perhaps all this belongs to `getProsemirrorSetup`?
      plugins: [
        ySyncPlugin(type),
        // There is also `prosemirror-history` module, but I believe
        // yUndoPlugin is supposed to be a complete replacement for it.
        // See https://codeberg.org/webxdc/editor/issues/13#issuecomment-1006670
        yUndoPlugin(),
        keymap({
          'Mod-z': undo,
          'Mod-y': redo,
          'Mod-Shift-z': redo,
          'Mod-s': () => {
            provider.sendUnsentLocalUpdates();
            return true;
          },
        }),
      ].concat(getProsemirrorSetup({
        schema,
        undo,
        redo,
      })),
    }),
  });
}

/** @type {() => void} */
let updateSyncButtonState = () => {};

function initExtraButtons() {
  const extraButtonsWrapper = /** @type {HTMLSpanElement} */
    (document.getElementById('extra-buttons'));
  const menuBar = document.getElementsByClassName('ProseMirror-menubar')[0];
  menuBar.appendChild(extraButtonsWrapper);
  extraButtonsWrapper.style.display = '';

  const syncButton = /** @type {HTMLButtonElement} */
    (document.getElementById('sync'));
  syncButton.addEventListener('click', () => {
    provider.sendUnsentLocalUpdates();
  });
  updateSyncButtonState = () => {
    syncButton.disabled = !provider.haveUnsentUpdates;
  }
  updateSyncButtonState();
}
